module ContentStore where

import Research
import Parser

import Control.Exception (catch, IOException)
import Control.Monad (foldM, forM)
import Control.Monad.IO.Class (liftIO)
import Crypto.Hash (hash, SHA256, Digest)
import Data.ByteString (ByteString)
import Data.Maybe (catMaybes)
import Data.Text (Text)
import Database.SQLite.Simple
import System.Directory (createDirectoryIfMissing, getXdgDirectory, XdgDirectory(..))
import System.FilePath ((</>), takeDirectory)

import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as LBS
import qualified Data.Map as Map
import qualified Data.Serialize as Cereal
import qualified Data.Text as T
import qualified Data.Text.Encoding as TE
import Database.SQLite.Simple.FromRow (FromRow(..), field)
import Network.HTTP.Simple (httpJSON, parseRequest, getResponseBody, getResponseStatus)
import Network.HTTP.Types.Status (statusCode)
import Data.Aeson (FromJSON(..), ToJSON(..), object, (.=), (.:), Value)
import qualified Data.Aeson as Aeson
import Network.Wai (Application, responseLBS, requestMethod, pathInfo, lazyRequestBody)
import Network.Wai.Handler.Warp (run)
import Network.HTTP.Types (status200, status404, status400, methodGet, methodPost)
import System.Environment (lookupEnv)
import Data.Time.Clock.POSIX (getPOSIXTime)
import Data.List (intercalate)

data StoredTerm = StoredTerm
  { termHash :: Text
  , termNames :: Text
  , termData :: ByteString
  , termMetadata :: Text
  , termCreatedAt :: Integer
  , termTags :: Text
  } deriving (Show)

instance FromRow StoredTerm where
  fromRow = StoredTerm <$> field <*> field <*> field <*> field <*> field <*> field

initContentStore :: IO Connection
initContentStore = do
  dbPath <- getContentStorePath
  createDirectoryIfMissing True (takeDirectory dbPath)
  conn <- open dbPath
  execute_ conn "CREATE TABLE IF NOT EXISTS terms (\
                \hash TEXT PRIMARY KEY, \
                \names TEXT, \
                \term_data BLOB, \
                \metadata TEXT, \
                \created_at INTEGER DEFAULT (strftime('%s','now')), \
                \tags TEXT DEFAULT '')"
  execute_ conn "CREATE INDEX IF NOT EXISTS terms_names_idx ON terms(names)"
  execute_ conn "CREATE INDEX IF NOT EXISTS terms_tags_idx ON terms(tags)"

  return conn

getContentStorePath :: IO FilePath
getContentStorePath = do
  dataDir <- getXdgDirectory XdgData "tricu"
  return $ dataDir </> "content-store.db"

instance Cereal.Serialize T where
  put Leaf = Cereal.putWord8 0
  put (Stem t) = do
    Cereal.putWord8 1
    Cereal.put t
  put (Fork a b) = do
    Cereal.putWord8 2
    Cereal.put a
    Cereal.put b

  get = do
    tag <- Cereal.getWord8
    case tag of
      0 -> return Leaf
      1 -> Stem <$> Cereal.get
      2 -> Fork <$> Cereal.get <*> Cereal.get
      _ -> fail $ "Invalid tag for T: " ++ show tag

serializeTerm :: T -> ByteString
serializeTerm = LBS.toStrict . Cereal.encodeLazy

deserializeTerm :: ByteString -> Either String T
deserializeTerm = Cereal.decodeLazy . LBS.fromStrict

hashTerm :: T -> Text
hashTerm = T.pack . show . (hash :: ByteString -> Digest SHA256) . serializeTerm

storeTerm :: Connection -> [String] -> T -> IO Text
storeTerm conn names term = do
  let termBS = serializeTerm term
      termHashText = hashTerm term
      namesText = T.pack $ intercalate "," names
      metadataText = T.pack "{}"
      tagsText = T.pack ""

  existing <- query conn
    "SELECT hash FROM terms WHERE hash = ?"
    (Only termHashText) :: IO [Only Text]

  case existing of
    [] -> execute conn
             "INSERT INTO terms (hash, names, term_data, metadata, tags) VALUES (?, ?, ?, ?, ?)"
             (termHashText, namesText, termBS, metadataText, tagsText)
    _  -> execute conn
             "UPDATE terms SET names = ?, metadata = ? WHERE hash = ?"
             (namesText, metadataText, termHashText)

  return termHashText

getTermByHash :: Connection -> Text -> IO (Maybe StoredTerm)
getTermByHash conn hashText = do
  results <- query conn
    "SELECT hash, names, term_data, metadata, created_at, tags FROM terms WHERE hash = ?"
    (Only hashText)
  case results of
    [term] -> return $ Just term
    _      -> return Nothing

getTermByName :: Connection -> Text -> IO (Maybe StoredTerm)
getTermByName conn nameText = do
  results <- query conn
    "SELECT hash, names, term_data, metadata, created_at, tags FROM terms WHERE names = ? ORDER BY created_at DESC LIMIT 1"
    (Only nameText)
  case results of
    [term] -> return $ Just term
    _      -> return Nothing

listStoredTerms :: Connection -> IO [StoredTerm]
listStoredTerms conn = do
  query_ conn "SELECT hash, names, term_data, metadata, created_at, tags FROM terms ORDER BY created_at DESC"

storeEnvironment :: Connection -> Env -> IO [(String, Text)]
storeEnvironment conn env = do
  let defs = Map.toList $ Map.delete "!result" env
  
  -- Group definitions by their tree form
  let groupedDefs = Map.toList $ Map.fromListWith (++) [(term, [name]) | (name, term) <- defs]
  
  -- Store each unique tree form with all its names
  forM groupedDefs $ \(term, namesList) -> do
    -- namesList is [String], which is what storeTerm expects
    hash <- storeTerm conn namesList term
    -- Return the first name and hash for backward compatibility
    return (head namesList, hash)

-- API types for remote term lookup
data TermResponse = TermResponse {
  respHash :: Text,
  respNames :: Text,
  respData :: ByteString,
  respMetadata :: Text,
  respCreatedAt :: Integer,
  respTags :: Text
}

instance FromJSON TermResponse where
  parseJSON = Aeson.withObject "TermResponse" $ \v -> TermResponse
    <$> v .: "hash"
    <*> v .: "names"
    <*> (TE.encodeUtf8 <$> v .: "data")
    <*> v .: "metadata"
    <*> v .: "created_at"
    <*> v .: "tags"

instance ToJSON StoredTerm where
  toJSON term = object [
      "hash" .= termHash term,
      "names" .= termNames term,
      "data" .= TE.decodeUtf8 (termData term),
      "metadata" .= termMetadata term,
      "created_at" .= termCreatedAt term,
      "tags" .= termTags term
    ]

instance FromJSON StoredTerm where
  parseJSON = Aeson.withObject "StoredTerm" $ \v -> StoredTerm
    <$> v .: "hash"
    <*> v .: "names"
    <*> (TE.encodeUtf8 <$> v .: "data")
    <*> v .: "metadata"
    <*> v .: "created_at"
    <*> v .: "tags"

-- Try to fetch a term from a remote server if it's not found locally
fetchRemoteTerm :: String -> String -> IO (Maybe StoredTerm)
fetchRemoteTerm serverUrl identifier = do
  let isFullHash = '#' `elem` identifier && length (drop 1 identifier) == 64
      isName = not ('#' `elem` identifier)
      
  if not (isFullHash || isName)
    then do
      putStrLn "Cannot fetch remote term: incomplete hash provided"
      return Nothing
    else do
      let url = if isFullHash
                then serverUrl ++ "/term/hash/" ++ drop 1 identifier
                else serverUrl ++ "/term/name/" ++ identifier
      
      request <- parseRequest url
      response <- httpJSON request `catch` \(_ :: IOException) -> do
        putStrLn $ "Failed to connect to remote server: " ++ serverUrl
        return undefined  -- This will be caught by the pattern match below
      
      case statusCode (getResponseStatus response) of
        200 -> do
          let termResp = getResponseBody response :: TermResponse
          return $ Just $ StoredTerm 
            (respHash termResp) 
            (respNames termResp) 
            (respData termResp) 
            (respMetadata termResp) 
            (respCreatedAt termResp)
            (respTags termResp)
        _ -> return Nothing

-- Modified loadTerm to try remote lookup if term not found locally
loadTerm :: Connection -> String -> IO (Maybe T)
loadTerm conn identifier = do
  result <- loadLocalTerm conn identifier
  case result of
    Just t -> return $ Just t
    Nothing -> do
      serverUrl <- lookupEnv "TRICU_SERVER"
      case serverUrl of
        Just url -> do
          putStrLn $ "Term not found locally, trying remote server: " ++ url
          remoteTerm <- fetchRemoteTerm url identifier
          case remoteTerm of
            Just term -> do
              -- Store the term locally for future use
              execute conn
                "INSERT INTO terms (hash, names, term_data, metadata, created_at, tags) VALUES (?, ?, ?, ?, ?, ?)"
                (termHash term, termNames term, termData term, termMetadata term, termCreatedAt term, termTags term)
              deserializeStoredTerm term
            Nothing -> return Nothing
        Nothing -> return Nothing
  where
    loadLocalTerm conn identifier = do
      if '#' `elem` identifier
        then do
          let hashText = T.pack $ drop 1 identifier
          storedTerm <- getTermByHash conn hashText
          case storedTerm of
            Just term -> deserializeStoredTerm term
            Nothing -> return Nothing
        else do
          storedTerm <- getTermByName conn (T.pack identifier)
          case storedTerm of
            Just term -> deserializeStoredTerm term
            Nothing -> return Nothing
    
    deserializeStoredTerm term =
      case deserializeTerm (termData term) of
        Right t -> return $ Just t
        Left err -> do
          putStrLn $ "Error deserializing term: " ++ err
          return Nothing

-- Server implementation
runServer :: Int -> IO ()
runServer port = do
  conn <- initContentStore
  putStrLn $ "Starting tricu server on port " ++ show port
  run port (app conn)

app :: Connection -> Application
app conn req respond = do
  case (requestMethod req, pathInfo req) of
    -- Get term by hash
    (method, ["term", "hash", hash]) | method == methodGet -> do
      result <- getTermByHash conn hash
      case result of
        Just term -> respond $ responseLBS status200 [("Content-Type", "application/json")] 
                              $ Aeson.encode term
        Nothing -> respond $ responseLBS status404 [("Content-Type", "application/json")] 
                            $ "{\"error\": \"Term not found\"}"
    
    -- Get term by name
    (method, ["term", "name", name]) | method == methodGet -> do
      result <- getTermByName conn name
      case result of
        Just term -> respond $ responseLBS status200 [("Content-Type", "application/json")] 
                              $ Aeson.encode term
        Nothing -> respond $ responseLBS status404 [("Content-Type", "application/json")] 
                            $ "{\"error\": \"Term not found\"}"
    
    -- Store a new term
    (method, ["term"]) | method == methodPost -> do
      body <- lazyRequestBody req
      case Aeson.decode body :: Maybe StoredTerm of
        Just term -> do
          -- Update the timestamp
          currentTime <- round <$> getPOSIXTime
          let updatedTerm = term { termCreatedAt = currentTime }
          
          execute conn
            "INSERT INTO terms (hash, names, term_data, metadata, created_at, tags) VALUES (?, ?, ?, ?, ?, ?)"
            (termHash updatedTerm, termNames updatedTerm, termData updatedTerm, termMetadata updatedTerm, termCreatedAt updatedTerm, termTags updatedTerm)
          
          respond $ responseLBS status200 [("Content-Type", "application/json")] 
                  $ "{\"status\": \"success\", \"hash\": \"" <> LBS.fromStrict (TE.encodeUtf8 (termHash updatedTerm)) <> "\"}"
        
        Nothing -> respond $ responseLBS status400 [("Content-Type", "application/json")] 
                            $ "{\"error\": \"Invalid request format\"}"
    
    -- List all terms
    (method, ["terms"]) | method == methodGet -> do
      terms <- listStoredTerms conn
      respond $ responseLBS status200 [("Content-Type", "application/json")] 
                $ Aeson.encode terms
    
    -- Get terms by tag
    (method, ["terms", "tag", tag]) | method == methodGet -> do
      terms <- getTermsByTag conn tag
      respond $ responseLBS status200 [("Content-Type", "application/json")] 
                $ Aeson.encode terms
    
    -- Default response for unknown routes
    _ -> respond $ responseLBS status404 [("Content-Type", "application/json")] 
                  $ "{\"error\": \"Not found\"}"

loadEnvironment :: Connection -> IO Env
loadEnvironment conn = do
  terms <- listStoredTerms conn
  foldM addTermToEnv Map.empty terms
  where
    addTermToEnv env term =
      case deserializeTerm (termData term) of
        Right t -> do
          -- Split the names and add each one to the environment
          let namesList = T.splitOn "," (termNames term)
          return $ foldl (\e name -> Map.insert (T.unpack name) t e) env namesList
        Left _ -> return env

getTermVersions :: Connection -> String -> IO [(Text, T, Integer)]
getTermVersions conn name = do
  let nameText = T.pack name
  results <- query conn
    "SELECT hash, term_data, created_at FROM terms WHERE names LIKE ? ORDER BY created_at DESC"
    (Only $ "%" <> nameText <> "%")

  catMaybes <$> mapM (\(hash, termData, timestamp) ->
    case deserializeTerm termData of
      Right t -> return $ Just (hash, t, timestamp)
      Left err -> do
        putStrLn $ "Error deserializing term: " ++ err
        return Nothing) results

-- Set a tag for a term
setTag :: Connection -> Text -> Text -> IO ()
setTag conn hash tagValue = do
  -- Check if the term exists
  exists <- termExists conn hash
  if exists
    then do
      -- Get current tags
      currentTags <- query conn "SELECT tags FROM terms WHERE hash = ?" (Only hash) :: IO [Only Text]
      case currentTags of
        [Only tags] -> do
          let tagsList = filter (not . T.null) $ T.splitOn "," tags
              newTags = if tagValue `elem` tagsList
                        then tags  -- Tag already exists
                        else if T.null tags
                             then tagValue
                             else tags <> "," <> tagValue
          -- Update tags
          execute conn "UPDATE terms SET tags = ? WHERE hash = ?" (newTags, hash)
        _ -> putStrLn $ "Term with hash " ++ T.unpack hash ++ " not found"
    else
      putStrLn $ "Term with hash " ++ T.unpack hash ++ " does not exist"

-- Check if a term exists
termExists :: Connection -> Text -> IO Bool
termExists conn hash = do
  results <- query conn "SELECT 1 FROM terms WHERE hash = ? LIMIT 1" (Only hash) :: IO [[Int]]
  return $ not (null results)

-- Get all tags for a term
getTagsForTerm :: Connection -> Text -> IO [Text]
getTagsForTerm conn hash = do
  tags <- query conn "SELECT tags FROM terms WHERE hash = ?" (Only hash) :: IO [Only Text]
  case tags of
    [Only tagsText] -> return $ filter (not . T.null) $ T.splitOn "," tagsText
    _ -> return []

-- Get all terms with a specific tag
getTermsByTag :: Connection -> Text -> IO [StoredTerm]
getTermsByTag conn tagValue = do
  -- Use LIKE with pattern matching to find terms with the tag
  -- We need to match ",tag," or "tag," or ",tag" or just "tag" if it's the only tag
  let pattern = "%," <> tagValue <> ",%"
      pattern2 = tagValue <> ",%"
      pattern3 = "%," <> tagValue
      exactMatch = tagValue
  query conn "SELECT hash, names, term_data, metadata, created_at, tags \
             \FROM terms \
             \WHERE tags LIKE ? OR tags LIKE ? OR tags LIKE ? OR tags = ? \
             \ORDER BY created_at DESC" 
             (pattern, pattern2, pattern3, exactMatch)

-- Get all terms with any tag
getTaggedTerms :: Connection -> IO [StoredTerm]
getTaggedTerms conn = do
  query_ conn "SELECT hash, names, term_data, metadata, created_at, tags \
              \FROM terms \
              \WHERE tags != '' \
              \ORDER BY created_at DESC"
