module Eval where

import ContentStore
import Parser
import Research

import Control.Monad (forM_, foldM)
import Data.List (partition, (\\))
import Data.Map  (Map)
import Database.SQLite.Simple
import qualified Data.Map as Map
import qualified Data.Set as Set
import qualified Data.Text as T
import Data.List (foldl')

evalSingle :: Env -> TricuAST -> Env
evalSingle env term
  | SDef name [] body <- term
  = case Map.lookup name env of
      Just existingValue
        | existingValue == evalASTSync env body -> env
        | otherwise
        -> let res = evalASTSync env body
           in Map.insert "!result" res (Map.insert name res env)
      Nothing
        -> let res = evalASTSync env body
           in Map.insert "!result" res (Map.insert name res env)
  | SApp func arg <- term
  = let res = apply (evalASTSync env func) (evalASTSync env arg)
    in Map.insert "!result" res env
  | SVar name Nothing <- term
  = case Map.lookup name env of
      Just v -> Map.insert "!result" v env
      Nothing -> errorWithoutStackTrace $ "Variable " ++ name ++ " not defined"
  | SVar name (Just hash) <- term
  = errorWithoutStackTrace $ "Hash-specific variable lookup not supported in local evaluation: " ++ name ++ "#" ++ hash
  | otherwise
  = let res = evalASTSync env term
    in Map.insert "!result" res env

evalTricu :: Env -> [TricuAST] -> Env
evalTricu env x = go env (reorderDefs env x)
  where
    go env []     = env
    go env [x]    =
      let updatedEnv = evalSingle env x
      in Map.insert "!result" (result updatedEnv) updatedEnv
    go env (x:xs) =
      evalTricu (evalSingle env x) xs

-- Pure evaluation function that doesn't depend on IO
evalASTSync :: Env -> TricuAST -> T
evalASTSync env term = case term of
  SLambda _ _ -> evalASTSync env (elimLambda term)
  SVar name Nothing -> case Map.lookup name env of
    Just v -> v
    Nothing -> errorWithoutStackTrace $ "Variable " ++ name ++ " not defined"
  SVar name (Just hash) ->
    -- In pure evaluation, we can only look up variables that are in the environment
    case Map.lookup (name ++ "#" ++ hash) env of
      Just v -> v
      Nothing -> errorWithoutStackTrace $
        "Variable " ++ name ++ " with hash " ++ hash ++ " not found in environment"
  TLeaf -> Leaf
  TStem t -> Stem (evalASTSync env t)
  TFork t u -> Fork (evalASTSync env t) (evalASTSync env u)
  SApp t u -> apply (evalASTSync env t) (evalASTSync env u)
  SStr s -> ofString s
  SInt n -> ofNumber n
  SList xs -> ofList (map (evalASTSync env) xs)
  SEmpty -> Leaf
  _ -> errorWithoutStackTrace $ "Unexpected AST term: " ++ show term

-- IO wrapper that resolves terms from the database before evaluation
evalAST :: Maybe Connection -> Map.Map String T.Text -> TricuAST -> IO T
evalAST mconn selectedVersions ast = do
  -- First, collect all variable names that need to be resolved
  let varNames = collectVarNames ast

  -- Resolve all needed variables from the database
  resolvedEnv <- resolveTermsFromStore mconn selectedVersions varNames

  -- Now perform pure evaluation with the resolved environment
  return $ evalASTSync resolvedEnv ast

-- Helper to collect all variable names from an AST
collectVarNames :: TricuAST -> [(String, Maybe String)]
collectVarNames = go []
  where
    go acc (SVar name mhash) = (name, mhash) : acc
    go acc (SApp t u) = go (go acc t) u
    go acc (SLambda vars body) =
      -- Filter out bound variables
      let boundVars = Set.fromList vars
          collected = go [] body
      in acc ++ filter (\(name, _) -> not $ Set.member name boundVars) collected
    go acc (TStem t) = go acc t
    go acc (TFork t u) = go (go acc t) u
    go acc (SList xs) = foldl' go acc xs
    go acc _ = acc

-- Resolve terms from the content store
resolveTermsFromStore :: Maybe Connection -> Map.Map String T.Text -> [(String, Maybe String)] -> IO Env
resolveTermsFromStore Nothing _ _ = return Map.empty
resolveTermsFromStore (Just conn) selectedVersions varNames = do
  -- Process each variable and build the environment
  foldM (\env (name, mhash) -> do
    term <- resolveTermFromStore conn selectedVersions name mhash
    case term of
      Just t -> return $ Map.insert (getVarKey name mhash) t env
      Nothing -> return env
      ) Map.empty varNames
  where
    getVarKey name Nothing = name
    getVarKey name (Just hash) = name ++ "#" ++ hash

-- Helper to resolve a single term from the store
resolveTermFromStore :: Connection -> Map.Map String T.Text -> String -> Maybe String -> IO (Maybe T)
resolveTermFromStore conn selectedVersions name mhash = case mhash of
  -- If a specific hash is provided in the code, use that
  Just hashPrefix -> do
    -- Find all terms with this name
    versions <- getTermVersions conn name

    -- Filter for versions that match the hash prefix
    let matchingVersions = filter (\(hash, _, _) ->
          T.isPrefixOf (T.pack hashPrefix) hash) versions

    case matchingVersions of
      [] -> return Nothing
      [(_, term, _)] -> return $ Just term
      _ -> return Nothing  -- Ambiguous hash prefix

  -- No hash specified, fall back to the selectedVersions or normal lookup
  Nothing -> case Map.lookup name selectedVersions of
    -- If we have a selected version, use that hash
    Just hash -> do
      mterm <- getTermByHash conn hash
      case mterm of
        Just term -> case deserializeTerm (termData term) of
          Right t -> return $ Just t
          Left _ -> return Nothing
        Nothing -> return Nothing

    -- Otherwise, try to load by name
    Nothing -> do
      versions <- getTermVersions conn name
      case versions of
        [] -> return Nothing
        [(_, term, _)] -> return $ Just term
        _ -> return $ Just $ (\(_, t, _) -> t) $ head versions  -- Use most recent version

elimLambda :: TricuAST -> TricuAST
elimLambda = go
  where
    go term
      | etaReduction term   = elimLambda $ etaReduceResult term
      | triagePattern term  = _TRI
      | composePattern term = _B
      | lambdaList term     = elimLambda $ lambdaListResult term
      | nestedLambda term   = nestedLambdaResult term
      | application term    = applicationResult term
      | otherwise           = term

    etaReduction (SLambda [v] (SApp f (SVar x Nothing))) = v == x && not (isFree v f)
    etaReduction _ = False
    etaReduceResult (SLambda [_] (SApp f _)) = f

    triagePattern (SLambda [a] (SLambda [b] (SLambda [c] body))) = body == triageBody a b c
    triagePattern _ = False

    composePattern (SLambda [f] (SLambda [g] (SLambda [x] body))) = body == composeBody f g x
    composePattern _ = False

    lambdaList (SLambda [_] (SList _)) = True
    lambdaList _ = False
    lambdaListResult (SLambda [v] (SList xs)) = SLambda [v] (foldr wrapTLeaf TLeaf xs)
    wrapTLeaf m r = SApp (SApp TLeaf m) r

    nestedLambda (SLambda (_:_) _) = True
    nestedLambda _ = False
    nestedLambdaResult (SLambda (v:vs) body)
      | null vs   = toSKI v (elimLambda body)
      | otherwise = elimLambda (SLambda [v] (SLambda vs body))

    application (SApp _ _) = True
    application _ = False
    applicationResult (SApp f g) = SApp (elimLambda f) (elimLambda g)

    toSKI x (SVar y Nothing)
      | x == y = _I
      | otherwise = SApp _K (SVar y Nothing)
    toSKI x (SApp m n) = SApp (SApp _S (toSKI x m)) (toSKI x n)
    toSKI x (SLambda [y] body) = toSKI x (toSKI y body)
    toSKI _ term = SApp _K term

    -- Combinators and special forms
    _S   = parseSingle "t (t (t t t)) t"
    _K   = parseSingle "t t"
    _I   = parseSingle "t (t (t t)) t"
    _B   = parseSingle "t (t (t t (t (t (t t t)) t))) (t t)"
    _TRI = parseSingle "t (t (t t (t (t (t t t))))) t"

    -- Pattern bodies
    triageBody a b c = SApp (SApp TLeaf (SApp (SApp TLeaf (SVar a Nothing)) (SVar b Nothing))) (SVar c Nothing)
    composeBody f g x = SApp (SVar f Nothing) (SVar g Nothing)

isFree :: String -> TricuAST -> Bool
isFree x = Set.member x . freeVars

freeVars :: TricuAST -> Set.Set String
freeVars (SVar v Nothing) = Set.singleton v
freeVars (SVar v (Just _)) = Set.singleton v  -- Hash doesn't affect free variables
freeVars (SApp t u) = Set.union (freeVars t) (freeVars u)
freeVars (SLambda vs body) = Set.difference (freeVars body) (Set.fromList vs)
freeVars _ = Set.empty

reorderDefs :: Env -> [TricuAST] -> [TricuAST]
reorderDefs env defs
  | not (null missingDeps) =
      errorWithoutStackTrace $
        "Missing dependencies detected: " ++ show missingDeps
  | otherwise = orderedDefs ++ others
  where
    (defsOnly, others) = partition isDef defs
    defNames = [ name | SDef name _ _ <- defsOnly ]

    defsWithFreeVars = [(def, freeVars body) | def@(SDef _ _ body) <- defsOnly]

    graph = buildDepGraph defsOnly
    sortedDefs = sortDeps graph
    defMap = Map.fromList [(name, def) | def@(SDef name _ _) <- defsOnly]
    orderedDefs = map (defMap Map.!) sortedDefs

    freeVarsDefs = foldMap snd defsWithFreeVars
    freeVarsOthers = foldMap freeVars others
    allFreeVars = freeVarsDefs <> freeVarsOthers
    validNames = Set.fromList defNames `Set.union` Set.fromList (Map.keys env)
    missingDeps = Set.toList (allFreeVars `Set.difference` validNames)

    isDef SDef {} = True
    isDef _       = False

buildDepGraph :: [TricuAST] -> Map.Map String (Set.Set String)
buildDepGraph topDefs
  | not (null conflictingDefs) =
      errorWithoutStackTrace $
        "Conflicting definitions detected: " ++ show conflictingDefs
  | otherwise =
      Map.fromList
        [ (name, depends topDefs (SDef name [] body))
        | SDef name _ body <- topDefs]
  where
    defsMap = Map.fromListWith (++)
      [(name, [(name, body)]) | SDef name _ body <- topDefs]

    conflictingDefs =
      [ name
      | (name, defs) <- Map.toList defsMap
      , let bodies = map snd defs
      , not $ all (== head bodies) (tail bodies)
      ]

sortDeps :: Map.Map String (Set.Set String) -> [String]
sortDeps graph = go [] Set.empty (Map.keys graph)
  where
    go sorted sortedSet [] = sorted
    go sorted sortedSet remaining =
      let ready = [ name | name <- remaining
                        , let deps = Map.findWithDefault Set.empty name graph
                        , Set.isSubsetOf deps sortedSet ]
          notReady = remaining \\ ready
      in if null ready
         then errorWithoutStackTrace
          "ERROR: Cyclic dependency detected and prohibited.\n\
          \RESOLVE: Use nested lambdas."
         else go (sorted ++ ready)
                 (Set.union sortedSet (Set.fromList ready))
                 notReady

depends :: [TricuAST] -> TricuAST -> Set.Set String
depends topDefs (SDef _ _ body) =
  Set.intersection
    (Set.fromList [n | SDef n _ _ <- topDefs])
    (freeVars body)
depends _ _ = Set.empty

result :: Env -> T
result r = case Map.lookup "!result" r of
  Just a -> a
  Nothing -> errorWithoutStackTrace "No !result field found in provided env"

mainResult :: Env -> T
mainResult r = case Map.lookup "main" r of
  Just  a -> a
  Nothing -> errorWithoutStackTrace "No valid definition for `main` found."

-- Evaluate an AST with an explicit environment
evalWithEnv :: Env -> Maybe Connection -> Map.Map String T.Text -> TricuAST -> IO T
evalWithEnv env mconn selectedVersions ast = do
  -- First collect all variable names that might need to be resolved
  let varNames = findVarNames ast

  -- For each name not in the environment, try to resolve it from the store
  resolvedEnv <- case mconn of
    Just conn -> foldM (\e name ->
      if Map.member name e
        then return e
        else do
          mterm <- resolveTermFromStore conn selectedVersions name Nothing
          case mterm of
            Just term -> return $ Map.insert name term e
            Nothing -> return e
      ) env varNames
    Nothing -> return env

  -- Now evaluate with the resolved environment
  return $ evalASTSync resolvedEnv ast

-- Helper to find all variable names in an AST
findVarNames :: TricuAST -> [String]
findVarNames ast = case ast of
  SVar name _ -> [name]
  SApp a b -> findVarNames a ++ findVarNames b
  SLambda args body -> findVarNames body \\ args
  SDef name args body -> name : (findVarNames body \\ args)
  _ -> []
